/*
Copyright (c) 2009, Nigel Batten.
Contactable at <firstname>.<lastname>@mail.com

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

	1.	Redistributions of source code must retain the above copyright
		notice, this list of conditions and the following disclaimer.
	2.	Redistributions in binary form must reproduce the above copyright
		notice, this list of conditions and the following disclaimer in the
		documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.

===============================================
*/

#ifndef _VIDEO_H__
#define _VIDEO_H__

#include "font.h"
#include "video_sync.h"
#include "video_render.h"
#include "video_serial.h"
#include "../common/delay_x.h"

#define VERSION "1.0.4"

// PAL format is assumed unless NTSC is defined.
// (i.e. comment out the following line if you want PAL output)
// #define NTSC

#define ROW_COUNT 25
#define COL_COUNT 38

// Scanline definitions:
// SCANLINE_FIRST_DISPLAY
// 	determines which scanline we're on when the V-Sync routine has finished.
// SCANLINE_LAST_DISPLAY
//	determines when the V-Sync routine should start
// SCANLINE_FIRST_TEXTSAFE
//	determines where the first line of text should start being drawn
//  This may need to be adjusted to ensure that the text area appears
//  vertically centred. It will depend on your TV to some degree.

#ifdef NTSC
	// NTSC has the whole Vertical blanking period at the start
	// of the field/frame:
	// scanline
	// 1-9    v-sync pulses
	// 10-20  'blanked' scanlines (with normal HSync)
	// 21     first displayable line
	// 262    last displayable line
	#define SCANLINE_FIRST_DISPLAY 10
	#define SCANLINE_LAST_DISPLAY 262
	// SCANLINE_FIRST_TEXTSAFE: Technically, it should be 29, but on
	// my monitor, that's the very first line that appears (and is
	// actually chopped off when used Picture-in-picture).
	#define SCANLINE_FIRST_TEXTSAFE 30
#else
	// PAL has the following scanlines:
	// 1-5		broad v-sync pulses and short-pulse post-equalisation
	// 6-23		'blanked' scanlines (with normal HSync)
	// 24		first displayable line
	// 52		top line of text
	// 281		last line of text (TODO: CHECK THIS)
	// 309		last visible scanline
	// 310-312	short-pulse pre-equalisation
	#define SCANLINE_FIRST_DISPLAY 6
	#define SCANLINE_LAST_DISPLAY 309
	#define SCANLINE_FIRST_TEXTSAFE 52
#endif


// times, cycles and clock settings for various points in a scanline...
#define CYCLES_PER_US		(F_CPU / 1000000)
#ifdef NTSC
	// This is not quite the correct NTSC timing, but it produces a
	// nicely divisible-by-two figure of 1016 cycles per scanline.
	#define TIME_SCANLINE		63.5
#else
	#define TIME_SCANLINE		64
#endif

#define TIME_HALFSCANLINE	(TIME_SCANLINE / 2)
#define TIME_HSYNC			4
#define TIME_SHORTSYNC		2
#ifdef NTSC
	// NTSC has a smaller front-porch
	#define TIME_DISPLAYSTART	11.5
#else
	#define TIME_DISPLAYSTART	12.5
#endif

#define CYCLES_SCANLINE		(TIME_SCANLINE * CYCLES_PER_US)
#define CYCLES_HALFSCANLINE (TIME_HALFSCANLINE * CYCLES_PER_US)
#define CYCLES_HSYNC		(TIME_HSYNC * CYCLES_PER_US)
#define CYCLES_SHORTSYNC	(TIME_SHORTSYNC * CYCLES_PER_US)
#define CYCLES_DISPLAYSTART	(TIME_DISPLAYSTART * CYCLES_PER_US)

#define CLOCK_SCANLINE		(CYCLES_SCANLINE - 1 )
#define CLOCK_HALFSCANLINE	(CYCLES_HALFSCANLINE - 1 )
#define CLOCK_HSYNC			(CYCLES_HSYNC - 1 )
#define CLOCK_SHORTSYNC		(TIME_SHORTSYNC - 1 )
#define CLOCK_DISPLAYSTART	(CYCLES_DISPLAYSTART - 1 )
#define CLOCK_LONGSYNC		((CYCLES_HALFSCANLINE - CYCLES_HSYNC) - 1 )


#define SIG_SYNC_PIN 1
#define SIG_PIXEL_PIN 3
#define SPI_SS 2
#define SPI_SCK 5
#define SPI_MISO 4

#define SYNC_ON (PORTB &= (~(1<<SIG_SYNC_PIN)))
#define SYNC_OFF (PORTB |= (1<<SIG_SYNC_PIN))
#define PIXEL_ENABLE  DDRB = ((1<<SIG_PIXEL_PIN)|(1<<SIG_SYNC_PIN))
#define PIXEL_DISABLE DDRB = ((1<<SIG_SYNC_PIN))

// Note: The OPT flags don't correspond directly to the option jumpers.
// The option jumpers are read at start-up, and some may be mapped to bits
// in the options byte.

#define OPT_LINEOVERFLOW 0
#define OPT_HIDECURSOR 1
#define OPT_BLOCKCURSOR 2
#define OPT_AUTO_LF 3
#define OPT_AUTO_CR 4
#define OPT_KEYCLICK 5 // not yet supported

#define ROWINFO_DISPLAY 0
#define ROWINFO_DOUBLE_HEIGHT 1
#define ROWINFO_DOUBLE_HEIGHT_BOTTOM 2
#define ROWINFO_DOUBLE_WIDTH 3


// definition of the CYCLES macro.
// This is to wrap the _delay_cycles() macro in some code that will cause the
// compiler/linker to fail if a negative parameter is supplied.
// The optimiser removes the call to the (non-existent) ERROR_NEGATIVE_DELAY function if the delay is >= 0
void ERROR_NEGATIVE_DELAY( void ) ;
#define CYCLES(n) do { if (n < 0) ERROR_NEGATIVE_DELAY(); else _delay_cycles(n); }while(0)



// Global variables that have been shunted into GPIO (or otherwise unused registers).
//extern uint8_t g_char;				// Most recently received character
//extern uint8_t g_current_row;			// Current cursor row position
//extern uint8_t g_current_col;			// Current cursor column position
//extern uint16_t g_current_scanline;	// Current scanline. NOTE: 16-bit!
#define g_char 			TWBR
#define g_current_row	TCNT0
#define g_current_col	EEDR
#define g_current_scanline _SFR_IO16(0x23) // This consumes OCR2 and TCNT2

extern uint8_t g_options;			// various options (note: not directly mapped to jumpers)
extern uint8_t g_row_scroll_offset; // Which row should be displayed at the top. This is to enable simple scrolling.
extern uint8_t g_framecount;		// used to blink the cursor. Simply wraps around after 256 frames

extern uint8_t g_display[ROW_COUNT][COL_COUNT] ;
extern uint8_t g_rowinfo[ROW_COUNT];


#endif
