/*
Copyright (c) 2009, Nigel Batten.
Contactable at <firstname>.<lastname>@mail.com

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

	1.	Redistributions of source code must retain the above copyright
		notice, this list of conditions and the following disclaimer.
	2.	Redistributions in binary form must reproduce the above copyright
		notice, this list of conditions and the following disclaimer in the
		documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.

===============================================
*/

#include <avr/io.h>
#include <avr/sleep.h>
#include "video.h"

sync_handler *g_sync_handler = sync_handler_field_header;

inline void sync_handler_display( void )
{
/*
	SYNC_ON has already been called.
	4us Sync (during which the char_rx is called)
	8us pause (during which the char_handler is called)
*/
	g_char_poll();
	CYCLES( CYCLES_HSYNC - (CYCLES_CHAR_POLL + CYCLES_SYNC_HANDLER_ENTRY ) ) ;
	SYNC_OFF ;
	g_char_process();
	
	// OLD CODE : sleep_cpu(); // now go to sleep until the Display Start interrupt fires

	// The cycle-accurate timing of the 'start display' point used to
	// be accomplished by being woken by timer match 'B' *interrupt*.
	// I've now swapped to using a tight loop to wait for the timer match B *flag*.
	// The jitter in the timing (there's a potential 0-3 cycle difference in the timing)
	// is then compensated for by reading the current (low byte) timer value
	// and delaying according to the bottom two bits.
	//
	// Note that the delays were achieved empirically.
	// It fixed all cases of deliberately induced jitter.
	
	// Q. Why go to all this trouble?
	// A. The time-overhead is lower (by at least 16 cycles)
	//
	// Q. *Why* is the time-overhead lower?
	// A.
	// Going to sleep, waking up, servicing the interrupt routine (no matter how short) and
	// finally returning to the code following the sleep command - it all takes time.
	// The tight loop followed by the adjustment code is simply quicker.
	
	while( !(TIFR & (1 << OCF1B))){} // wait until the display start point is triggered.
	TIFR |= (1<< OCF1B) ; // clear the timer's flag
	
	// recover from jitter.
	asm("in		r0, %[_TCNT1L]		; read the timer counter					\n\t"
		"sbrs	r0, 0				; skip the following delay if bit 0 is set	\n\t"
		"rjmp	.+0					; 2 cycle nop								\n\t"
		"sbrc	r0, 1				; skip the following delay if bit 1 is clear\n\t"
		"lpm						; 3 cycle nop (trashes r0)					\n\t"
		:
		:
		[_TCNT1L]			"I" (_SFR_IO_ADDR(TCNT1L))
		:
		"r0");
}

inline void sync_handler_field_header( void )
{
/*
	SYNC_ON has already been called.
	
	PAL:
	This routine will run for 8 scanlines, with the following half-scanline pattern:
	6 short pulses
	5 broad pulses
	5 short pulses
	
	NTSC:
	This routine will run for 9 scanlines, with the following half-scanline pattern:
	6 short pulses
	6 broad pulses
	6 short pulses
*/
	// 2 PUSH instructions on entry, hence 4c fewer delay cycles required.
	CYCLES( CYCLES_SHORTSYNC - (CYCLES_SYNC_HANDLER_ENTRY + 4) ) ;
	SYNC_OFF ;
	
	OCR1A = CLOCK_HALFSCANLINE ;
	OCR1B = CLOCK_LONGSYNC ;
	g_char_poll();
	g_char_process();
	
	TIMSK |= (1<<OCIE1B) ; // enable the second interrupt.
	
	uint8_t i = 5 ; // 6 shorts in total (one has already been done).
	while( i-- )
	{
		sleep_cpu(); // will wake up at the longsync point. (not used in shortsync halflines)
		sleep_cpu(); // will wake up at the next half/whole scanline boundary.
		SYNC_ON ;
		CYCLES( CYCLES_SHORTSYNC - ( 2 ) ) ;
		SYNC_OFF ;
		g_char_poll() ;
		g_char_process();
	}
	
	sleep_cpu(); // will wake up at the longsync point (at the end of the last shortsync halfline)
	
#ifdef NTSC
	i = 6 ;
#else
	i = 5 ;
#endif
	while( i-- )
	{
		sleep_cpu(); // will wake up at the start of a whole/half scanline boundary
		SYNC_ON ;
		g_char_poll();
		g_char_process();
		sleep_cpu(); // will wake up at the longsync point
		SYNC_OFF ;
	}

#ifdef NTSC	
	i = 5 ; // only 5 in the loop. The extra one is done afterwards whilst changing back to full-scanline timing.
#else
	i = 4 ; // only 4 in the loop. The extra one is done afterwards whilst changing back to full-scanline timing.
#endif
	
	while( i-- )
	{
		sleep_cpu(); // will wake up at the whole/half scanline boundary
		SYNC_ON ;
		CYCLES( CYCLES_SHORTSYNC - (2)) ;
		SYNC_OFF ;
		g_char_poll();
		g_char_process();
		sleep_cpu(); // will wake up at the longsync point (not used in shortsync halflines)
	}
	
	// we are now just before the start of the last shortsync halfline...
	OCR1A = CLOCK_SCANLINE ; // reset back to normal.
	// there are 6 clock cycles since we last woke up.
	CYCLES(CYCLES_HSYNC - (6)); // we're now at the start of the last half-line.
	SYNC_ON;
	CYCLES(CYCLES_SHORTSYNC-(2));
	SYNC_OFF;
	OCR1B = CLOCK_DISPLAYSTART ; // restore the normal display interrupt.
	TIMSK &= ~(1<<OCIE1B) ; // disable the second interrupt.
	
	
	// no need to call the f_char_rx or g_char_handler routines in this last half-scanline.
	// - they only need calling once per whole scanline.

	g_sync_handler = sync_handler_display ;
	g_render_handler = render_handler_nondisplay ;
	g_current_scanline = SCANLINE_FIRST_DISPLAY - 1 ;
}
